@@ -5,13 +5,14 @@ from django.contrib.admin.models import LogEntry
 from django.contrib.admin.options import IncorrectLookupParameters
 from django.contrib.admin.templatetags.admin_list import pagination
 from django.contrib.admin.tests import AdminSeleniumTestCase
-from django.contrib.admin.views.main import ALL_VAR, SEARCH_VAR
+from django.contrib.admin.views.main import (
+    ALL_VAR, IS_POPUP_VAR, ORDER_VAR, PAGE_VAR, SEARCH_VAR, TO_FIELD_VAR,
+)
 from django.contrib.auth.models import User
 from django.contrib.contenttypes.models import ContentType
 from django.contrib.messages.storage.cookie import CookieStorage
 from django.db import connection, models
-from django.db.models import F
-from django.db.models.fields import Field, IntegerField
+from django.db.models import F, Field, IntegerField
 from django.db.models.functions import Upper
 from django.db.models.lookups import Contains, Exact
 from django.template import Context, Template, TemplateSyntaxError
@@ -41,7 +42,7 @@ from .models import (
 
 def build_tbody_html(pk, href, extra_fields):
     return (
-        '<tbody><tr class="row1">'
+        '<tbody><tr>'
         '<td class="action-checkbox">'
         '<input type="checkbox" name="_selected_action" value="{}" '
         'class="action-select"></td>'
@@ -258,7 +259,7 @@ class ChangeListTests(TestCase):
         Regression test for #14312: list_editable with pagination
         """
         new_parent = Parent.objects.create(name='parent')
-        for i in range(200):
+        for i in range(1, 201):
             Child.objects.create(name='name %s' % i, parent=new_parent)
         request = self.factory.get('/child/', data={'p': -1})  # Anything outside range
         request.user = self.superuser
@@ -273,7 +274,7 @@ class ChangeListTests(TestCase):
 
     def test_custom_paginator(self):
         new_parent = Parent.objects.create(name='parent')
-        for i in range(200):
+        for i in range(1, 201):
             Child.objects.create(name='name %s' % i, parent=new_parent)
 
         request = self.factory.get('/child/')
@@ -284,7 +285,7 @@ class ChangeListTests(TestCase):
         cl.get_results(request)
         self.assertIsInstance(cl.paginator, CustomPaginator)
 
-    def test_distinct_for_m2m_in_list_filter(self):
+    def test_no_duplicates_for_m2m_in_list_filter(self):
         """
         Regression test for #13902: When using a ManyToMany in list_filter,
         results shouldn't appear more than once. Basic ManyToMany.
@@ -304,8 +305,12 @@ class ChangeListTests(TestCase):
 
         # There's only one Group instance
         self.assertEqual(cl.result_count, 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_through_m2m_in_list_filter(self):
+    def test_no_duplicates_for_through_m2m_in_list_filter(self):
         """
         Regression test for #13902: When using a ManyToMany in list_filter,
         results shouldn't appear more than once. With an intermediate model.
@@ -324,12 +329,15 @@ class ChangeListTests(TestCase):
 
         # There's only one Group instance
         self.assertEqual(cl.result_count, 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_through_m2m_at_second_level_in_list_filter(self):
+    def test_no_duplicates_for_through_m2m_at_second_level_in_list_filter(self):
         """
         When using a ManyToMany in list_filter at the second level behind a
-        ForeignKey, distinct() must be called and results shouldn't appear more
-        than once.
+        ForeignKey, results shouldn't appear more than once.
         """
         lead = Musician.objects.create(name='Vox')
         band = Group.objects.create(name='The Hype')
@@ -346,8 +354,12 @@ class ChangeListTests(TestCase):
 
         # There's only one Concert instance
         self.assertEqual(cl.result_count, 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_inherited_m2m_in_list_filter(self):
+    def test_no_duplicates_for_inherited_m2m_in_list_filter(self):
         """
         Regression test for #13902: When using a ManyToMany in list_filter,
         results shouldn't appear more than once. Model managed in the
@@ -367,8 +379,12 @@ class ChangeListTests(TestCase):
 
         # There's only one Quartet instance
         self.assertEqual(cl.result_count, 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_m2m_to_inherited_in_list_filter(self):
+    def test_no_duplicates_for_m2m_to_inherited_in_list_filter(self):
         """
         Regression test for #13902: When using a ManyToMany in list_filter,
         results shouldn't appear more than once. Target of the relationship
@@ -388,11 +404,15 @@ class ChangeListTests(TestCase):
 
         # There's only one ChordsBand instance
         self.assertEqual(cl.result_count, 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_non_unique_related_object_in_list_filter(self):
+    def test_no_duplicates_for_non_unique_related_object_in_list_filter(self):
         """
-        Regressions tests for #15819: If a field listed in list_filters
-        is a non-unique related object, distinct() must be called.
+        Regressions tests for #15819: If a field listed in list_filters is a
+        non-unique related object, results shouldn't appear more than once.
         """
         parent = Parent.objects.create(name='Mary')
         # Two children with the same name
@@ -404,8 +424,12 @@ class ChangeListTests(TestCase):
         request.user = self.superuser
 
         cl = m.get_changelist_instance(request)
-        # Make sure distinct() was called
+        # Exists() is applied.
         self.assertEqual(cl.queryset.count(), 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
     def test_changelist_search_form_validation(self):
         m = ConcertAdmin(Concert, custom_site)
@@ -423,10 +447,10 @@ class ChangeListTests(TestCase):
                 self.assertEqual(1, len(messages))
                 self.assertEqual(error, messages[0])
 
-    def test_distinct_for_non_unique_related_object_in_search_fields(self):
+    def test_no_duplicates_for_non_unique_related_object_in_search_fields(self):
         """
         Regressions tests for #15819: If a field listed in search_fields
-        is a non-unique related object, distinct() must be called.
+        is a non-unique related object, Exists() must be applied.
         """
         parent = Parent.objects.create(name='Mary')
         Child.objects.create(parent=parent, name='Danielle')
@@ -437,13 +461,17 @@ class ChangeListTests(TestCase):
         request.user = self.superuser
 
         cl = m.get_changelist_instance(request)
-        # Make sure distinct() was called
+        # Exists() is applied.
         self.assertEqual(cl.queryset.count(), 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
-    def test_distinct_for_many_to_many_at_second_level_in_search_fields(self):
+    def test_no_duplicates_for_many_to_many_at_second_level_in_search_fields(self):
         """
         When using a ManyToMany in search_fields at the second level behind a
-        ForeignKey, distinct() must be called and results shouldn't appear more
+        ForeignKey, Exists() must be applied and results shouldn't appear more
         than once.
         """
         lead = Musician.objects.create(name='Vox')
@@ -459,6 +487,10 @@ class ChangeListTests(TestCase):
         cl = m.get_changelist_instance(request)
         # There's only one Concert instance
         self.assertEqual(cl.queryset.count(), 1)
+        # Queryset must be deletable.
+        self.assertIs(cl.queryset.query.distinct, False)
+        cl.queryset.delete()
+        self.assertEqual(cl.queryset.count(), 0)
 
     def test_pk_in_search_fields(self):
         band = Group.objects.create(name='The Hype')
@@ -551,23 +583,23 @@ class ChangeListTests(TestCase):
         cl = m.get_changelist_instance(request)
         self.assertCountEqual(cl.queryset, [abcd])
 
-    def test_no_distinct_for_m2m_in_list_filter_without_params(self):
+    def test_no_exists_for_m2m_in_list_filter_without_params(self):
         """
         If a ManyToManyField is in list_filter but isn't in any lookup params,
-        the changelist's query shouldn't have distinct.
+        the changelist's query shouldn't have Exists().
         """
         m = BandAdmin(Band, custom_site)
         for lookup_params in ({}, {'name': 'test'}):
             request = self.factory.get('/band/', lookup_params)
             request.user = self.superuser
             cl = m.get_changelist_instance(request)
-            self.assertFalse(cl.queryset.query.distinct)
+            self.assertNotIn(' EXISTS', str(cl.queryset.query))
 
-        # A ManyToManyField in params does have distinct applied.
+        # A ManyToManyField in params does have Exists() applied.
         request = self.factory.get('/band/', {'genres': '0'})
         request.user = self.superuser
         cl = m.get_changelist_instance(request)
-        self.assertTrue(cl.queryset.query.distinct)
+        self.assertIn(' EXISTS', str(cl.queryset.query))
 
     def test_pagination(self):
         """
@@ -575,7 +607,7 @@ class ChangeListTests(TestCase):
         use queryset set by modeladmin.
         """
         parent = Parent.objects.create(name='anything')
-        for i in range(30):
+        for i in range(1, 31):
             Child.objects.create(name='name %s' % i, parent=parent)
             Child.objects.create(name='filtered %s' % i, parent=parent)
 
@@ -651,7 +683,7 @@ class ChangeListTests(TestCase):
 
     def test_show_all(self):
         parent = Parent.objects.create(name='anything')
-        for i in range(30):
+        for i in range(1, 31):
             Child.objects.create(name='name %s' % i, parent=parent)
             Child.objects.create(name='filtered %s' % i, parent=parent)
 
@@ -707,6 +739,68 @@ class ChangeListTests(TestCase):
         link = reverse('admin:admin_changelist_parent_change', args=(p.pk,))
         self.assertNotContains(response, '<a href="%s">' % link)
 
+    def test_clear_all_filters_link(self):
+        self.client.force_login(self.superuser)
+        url = reverse('admin:auth_user_changelist')
+        response = self.client.get(url)
+        self.assertNotContains(response, '&#10006; Clear all filters')
+        link = '<a href="%s">&#10006; Clear all filters</a>'
+        for data, href in (
+            ({'is_staff__exact': '0'}, '?'),
+            (
+                {'is_staff__exact': '0', 'username__startswith': 'test'},
+                '?username__startswith=test',
+            ),
+            (
+                {'is_staff__exact': '0', SEARCH_VAR: 'test'},
+                '?%s=test' % SEARCH_VAR,
+            ),
+            (
+                {'is_staff__exact': '0', IS_POPUP_VAR: 'id'},
+                '?%s=id' % IS_POPUP_VAR,
+            ),
+        ):
+            with self.subTest(data=data):
+                response = self.client.get(url, data=data)
+                self.assertContains(response, link % href)
+
+    def test_clear_all_filters_link_callable_filter(self):
+        self.client.force_login(self.superuser)
+        url = reverse('admin:admin_changelist_band_changelist')
+        response = self.client.get(url)
+        self.assertNotContains(response, '&#10006; Clear all filters')
+        link = '<a href="%s">&#10006; Clear all filters</a>'
+        for data, href in (
+            ({'nr_of_members_partition': '5'}, '?'),
+            (
+                {'nr_of_members_partition': 'more', 'name__startswith': 'test'},
+                '?name__startswith=test',
+            ),
+            (
+                {'nr_of_members_partition': '5', IS_POPUP_VAR: 'id'},
+                '?%s=id' % IS_POPUP_VAR,
+            ),
+        ):
+            with self.subTest(data=data):
+                response = self.client.get(url, data=data)
+                self.assertContains(response, link % href)
+
+    def test_no_clear_all_filters_link(self):
+        self.client.force_login(self.superuser)
+        url = reverse('admin:auth_user_changelist')
+        link = '>&#10006; Clear all filters</a>'
+        for data in (
+            {SEARCH_VAR: 'test'},
+            {ORDER_VAR: '-1'},
+            {TO_FIELD_VAR: 'id'},
+            {PAGE_VAR: '1'},
+            {IS_POPUP_VAR: '1'},
+            {'username__startswith': 'test'},
+        ):
+            with self.subTest(data=data):
+                response = self.client.get(url, data=data)
+                self.assertNotContains(response, link)
+
     def test_tuple_list_display(self):
         swallow = Swallow.objects.create(origin='Africa', load='12.34', speed='22.2')
         swallow2 = Swallow.objects.create(origin='Africa', load='12.34', speed='22.2')
@@ -906,7 +1000,7 @@ class ChangeListTests(TestCase):
             custom_site.register(UnorderedObject, UnorderedObjectAdmin)
             model_admin = UnorderedObjectAdmin(UnorderedObject, custom_site)
             counter = 0 if ascending else 51
-            for page in range(0, 5):
+            for page in range(1, 6):
                 request = self._mocked_authenticated_request('/unorderedobject/?p=%s' % page, superuser)
                 response = model_admin.changelist_view(request)
                 for result in response.context_data['cl'].result_list:
@@ -950,7 +1044,7 @@ class ChangeListTests(TestCase):
             custom_site.register(OrderedObject, OrderedObjectAdmin)
             model_admin = OrderedObjectAdmin(OrderedObject, custom_site)
             counter = 0 if ascending else 51
-            for page in range(0, 5):
+            for page in range(1, 6):
                 request = self._mocked_authenticated_request('/orderedobject/?p=%s' % page, superuser)
                 response = model_admin.changelist_view(request)
                 for result in response.context_data['cl'].result_list:
@@ -1028,10 +1122,6 @@ class ChangeListTests(TestCase):
             (['field', '-other_field'], ['field', '-other_field']),
             # Composite unique nullable.
             (['-field', 'null_field'], ['-field', 'null_field', '-pk']),
-            # Composite unique nullable.
-            (['-field', 'null_field'], ['-field', 'null_field', '-pk']),
-            # Composite unique nullable.
-            (['-field', 'null_field'], ['-field', 'null_field', '-pk']),
             # Composite unique and nullable.
             (['-field', 'null_field', 'other_field'], ['-field', 'null_field', 'other_field']),
             # Composite unique attnames.
@@ -1051,6 +1141,98 @@ class ChangeListTests(TestCase):
             with self.subTest(ordering=ordering):
                 self.assertEqual(change_list._get_deterministic_ordering(ordering), expected)
 
+    @isolate_apps('admin_changelist')
+    def test_total_ordering_optimization_meta_constraints(self):
+        class Related(models.Model):
+            unique_field = models.BooleanField(unique=True)
+
+            class Meta:
+                ordering = ('unique_field',)
+
+        class Model(models.Model):
+            field_1 = models.BooleanField()
+            field_2 = models.BooleanField()
+            field_3 = models.BooleanField()
+            field_4 = models.BooleanField()
+            field_5 = models.BooleanField()
+            field_6 = models.BooleanField()
+            nullable_1 = models.BooleanField(null=True)
+            nullable_2 = models.BooleanField(null=True)
+            related_1 = models.ForeignKey(Related, models.CASCADE)
+            related_2 = models.ForeignKey(Related, models.CASCADE)
+            related_3 = models.ForeignKey(Related, models.CASCADE)
+            related_4 = models.ForeignKey(Related, models.CASCADE)
+
+            class Meta:
+                constraints = [
+                    *[
+                        models.UniqueConstraint(fields=fields, name=''.join(fields))
+                        for fields in (
+                            ['field_1'],
+                            ['nullable_1'],
+                            ['related_1'],
+                            ['related_2_id'],
+                            ['field_2', 'field_3'],
+                            ['field_2', 'nullable_2'],
+                            ['field_2', 'related_3'],
+                            ['field_3', 'related_4_id'],
+                        )
+                    ],
+                    models.CheckConstraint(check=models.Q(id__gt=0), name='foo'),
+                    models.UniqueConstraint(
+                        fields=['field_5'],
+                        condition=models.Q(id__gt=10),
+                        name='total_ordering_1',
+                    ),
+                    models.UniqueConstraint(
+                        fields=['field_6'],
+                        condition=models.Q(),
+                        name='total_ordering',
+                    ),
+                ]
+
+        class ModelAdmin(admin.ModelAdmin):
+            def get_queryset(self, request):
+                return Model.objects.none()
+
+        request = self._mocked_authenticated_request('/', self.superuser)
+        site = admin.AdminSite(name='admin')
+        model_admin = ModelAdmin(Model, site)
+        change_list = model_admin.get_changelist_instance(request)
+        tests = (
+            # Unique non-nullable field.
+            (['field_1'], ['field_1']),
+            # Unique nullable field.
+            (['nullable_1'], ['nullable_1', '-pk']),
+            # Related attname unique.
+            (['related_1_id'], ['related_1_id']),
+            (['related_2_id'], ['related_2_id']),
+            # Related ordering introspection is not implemented.
+            (['related_1'], ['related_1', '-pk']),
+            # Composite unique.
+            (['-field_2', 'field_3'], ['-field_2', 'field_3']),
+            # Composite unique nullable.
+            (['field_2', '-nullable_2'], ['field_2', '-nullable_2', '-pk']),
+            # Composite unique and nullable.
+            (
+                ['field_2', '-nullable_2', 'field_3'],
+                ['field_2', '-nullable_2', 'field_3'],
+            ),
+            # Composite field and related field name.
+            (['field_2', '-related_3'], ['field_2', '-related_3', '-pk']),
+            (['field_3', 'related_4'], ['field_3', 'related_4', '-pk']),
+            # Composite field and related field attname.
+            (['field_2', 'related_3_id'], ['field_2', 'related_3_id']),
+            (['field_3', '-related_4_id'], ['field_3', '-related_4_id']),
+            # Partial unique constraint is ignored.
+            (['field_5'], ['field_5', '-pk']),
+            # Unique constraint with an empty condition.
+            (['field_6'], ['field_6']),
+        )
+        for ordering, expected in tests:
+            with self.subTest(ordering=ordering):
+                self.assertEqual(change_list._get_deterministic_ordering(ordering), expected)
+
     def test_dynamic_list_filter(self):
         """
         Regression tests for ticket #17646: dynamic list_filter support.
@@ -1091,26 +1273,32 @@ class ChangeListTests(TestCase):
         request = self.factory.get('/group/')
         request.user = self.superuser
         cl = m.get_changelist_instance(request)
-        per_page = cl.list_per_page = 10
+        cl.list_per_page = 10
 
-        for page_num, objects_count, expected_page_range in [
-            (0, per_page, []),
-            (0, per_page * 2, list(range(2))),
-            (5, per_page * 11, list(range(11))),
-            (5, per_page * 12, [0, 1, 2, 3, 4, 5, 6, 7, 8, '.', 10, 11]),
-            (6, per_page * 12, [0, 1, '.', 3, 4, 5, 6, 7, 8, 9, 10, 11]),
-            (6, per_page * 13, [0, 1, '.', 3, 4, 5, 6, 7, 8, 9, '.', 11, 12]),
+        ELLIPSIS = cl.paginator.ELLIPSIS
+        for number, pages, expected in [
+            (1, 1, []),
+            (1, 2, [1, 2]),
+            (6, 11, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
+            (6, 12, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]),
+            (6, 13, [1, 2, 3, 4, 5, 6, 7, 8, 9, ELLIPSIS, 12, 13]),
+            (7, 12, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]),
+            (7, 13, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]),
+            (7, 14, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ELLIPSIS, 13, 14]),
+            (8, 13, [1, 2, ELLIPSIS, 5, 6, 7, 8, 9, 10, 11, 12, 13]),
+            (8, 14, [1, 2, ELLIPSIS, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]),
+            (8, 15, [1, 2, ELLIPSIS, 5, 6, 7, 8, 9, 10, 11, ELLIPSIS, 14, 15]),
         ]:
-            # assuming we have exactly `objects_count` objects
-            Group.objects.all().delete()
-            for i in range(objects_count):
-                Group.objects.create(name='test band')
+            with self.subTest(number=number, pages=pages):
+                # assuming exactly `pages * cl.list_per_page` objects
+                Group.objects.all().delete()
+                for i in range(pages * cl.list_per_page):
+                    Group.objects.create(name='test band')
 
-            # setting page number and calculating page range
-            cl.page_num = page_num
-            cl.get_results(request)
-            real_page_range = pagination(cl)['page_range']
-            self.assertEqual(expected_page_range, list(real_page_range))
+                # setting page number and calculating page range
+                cl.page_num = number
+                cl.get_results(request)
+                self.assertEqual(list(pagination(cl)['page_range']), expected)
 
     def test_object_tools_displayed_no_add_permission(self):
         """
@@ -1194,16 +1382,170 @@ class SeleniumTests(AdminSeleniumTestCase):
 
         # Test amount of rows in the Changelist
         rows = self.selenium.find_elements_by_css_selector(
-            '%s #result_list tbody tr' % form_id)
+            '%s #result_list tbody tr' % form_id
+        )
         self.assertEqual(len(rows), 1)
+        row = rows[0]
+
+        selection_indicator = self.selenium.find_element_by_css_selector(
+            '%s .action-counter' % form_id
+        )
+        all_selector = self.selenium.find_element_by_id('action-toggle')
+        row_selector = self.selenium.find_element_by_css_selector(
+            '%s #result_list tbody tr:first-child .action-select' % form_id
+        )
 
         # Test current selection
-        selection_indicator = self.selenium.find_element_by_css_selector(
-            '%s .action-counter' % form_id)
         self.assertEqual(selection_indicator.text, "0 of 1 selected")
+        self.assertIs(all_selector.get_property('checked'), False)
+        self.assertEqual(row.get_attribute('class'), '')
 
         # Select a row and check again
-        row_selector = self.selenium.find_element_by_css_selector(
-            '%s #result_list tbody tr:first-child .action-select' % form_id)
         row_selector.click()
         self.assertEqual(selection_indicator.text, "1 of 1 selected")
+        self.assertIs(all_selector.get_property('checked'), True)
+        self.assertEqual(row.get_attribute('class'), 'selected')
+
+        # Deselect a row and check again
+        row_selector.click()
+        self.assertEqual(selection_indicator.text, "0 of 1 selected")
+        self.assertIs(all_selector.get_property('checked'), False)
+        self.assertEqual(row.get_attribute('class'), '')
+
+    def test_modifier_allows_multiple_section(self):
+        """
+        Selecting a row and then selecting another row whilst holding shift
+        should select all rows in-between.
+        """
+        from selenium.webdriver.common.action_chains import ActionChains
+        from selenium.webdriver.common.keys import Keys
+
+        Parent.objects.bulk_create([Parent(name='parent%d' % i) for i in range(5)])
+        self.admin_login(username='super', password='secret')
+        self.selenium.get(self.live_server_url + reverse('admin:admin_changelist_parent_changelist'))
+        checkboxes = self.selenium.find_elements_by_css_selector('tr input.action-select')
+        self.assertEqual(len(checkboxes), 5)
+        for c in checkboxes:
+            self.assertIs(c.get_property('checked'), False)
+        # Check first row. Hold-shift and check next-to-last row.
+        checkboxes[0].click()
+        ActionChains(self.selenium).key_down(Keys.SHIFT).click(checkboxes[-2]).key_up(Keys.SHIFT).perform()
+        for c in checkboxes[:-2]:
+            self.assertIs(c.get_property('checked'), True)
+        self.assertIs(checkboxes[-1].get_property('checked'), False)
+
+    def test_select_all_across_pages(self):
+        Parent.objects.bulk_create([Parent(name='parent%d' % i) for i in range(101)])
+        self.admin_login(username='super', password='secret')
+        self.selenium.get(self.live_server_url + reverse('admin:admin_changelist_parent_changelist'))
+
+        selection_indicator = self.selenium.find_element_by_css_selector('.action-counter')
+        select_all_indicator = self.selenium.find_element_by_css_selector('.actions .all')
+        question = self.selenium.find_element_by_css_selector('.actions > .question')
+        clear = self.selenium.find_element_by_css_selector('.actions > .clear')
+        select_all = self.selenium.find_element_by_id('action-toggle')
+        select_across = self.selenium.find_element_by_name('select_across')
+
+        self.assertIs(question.is_displayed(), False)
+        self.assertIs(clear.is_displayed(), False)
+        self.assertIs(select_all.get_property('checked'), False)
+        self.assertEqual(select_across.get_property('value'), '0')
+        self.assertIs(selection_indicator.is_displayed(), True)
+        self.assertEqual(selection_indicator.text, '0 of 100 selected')
+        self.assertIs(select_all_indicator.is_displayed(), False)
+
+        select_all.click()
+        self.assertIs(question.is_displayed(), True)
+        self.assertIs(clear.is_displayed(), False)
+        self.assertIs(select_all.get_property('checked'), True)
+        self.assertEqual(select_across.get_property('value'), '0')
+        self.assertIs(selection_indicator.is_displayed(), True)
+        self.assertEqual(selection_indicator.text, '100 of 100 selected')
+        self.assertIs(select_all_indicator.is_displayed(), False)
+
+        question.click()
+        self.assertIs(question.is_displayed(), False)
+        self.assertIs(clear.is_displayed(), True)
+        self.assertIs(select_all.get_property('checked'), True)
+        self.assertEqual(select_across.get_property('value'), '1')
+        self.assertIs(selection_indicator.is_displayed(), False)
+        self.assertIs(select_all_indicator.is_displayed(), True)
+
+        clear.click()
+        self.assertIs(question.is_displayed(), False)
+        self.assertIs(clear.is_displayed(), False)
+        self.assertIs(select_all.get_property('checked'), False)
+        self.assertEqual(select_across.get_property('value'), '0')
+        self.assertIs(selection_indicator.is_displayed(), True)
+        self.assertEqual(selection_indicator.text, '0 of 100 selected')
+        self.assertIs(select_all_indicator.is_displayed(), False)
+
+    def test_actions_warn_on_pending_edits(self):
+        Parent.objects.create(name='foo')
+
+        self.admin_login(username='super', password='secret')
+        self.selenium.get(self.live_server_url + reverse('admin:admin_changelist_parent_changelist'))
+
+        name_input = self.selenium.find_element_by_id('id_form-0-name')
+        name_input.clear()
+        name_input.send_keys('bar')
+        self.selenium.find_element_by_id('action-toggle').click()
+        self.selenium.find_element_by_name('index').click()  # Go
+        alert = self.selenium.switch_to.alert
+        try:
+            self.assertEqual(
+                alert.text,
+                'You have unsaved changes on individual editable fields. If you '
+                'run an action, your unsaved changes will be lost.'
+            )
+        finally:
+            alert.dismiss()
+
+    def test_save_with_changes_warns_on_pending_action(self):
+        from selenium.webdriver.support.ui import Select
+
+        Parent.objects.create(name='parent')
+
+        self.admin_login(username='super', password='secret')
+        self.selenium.get(self.live_server_url + reverse('admin:admin_changelist_parent_changelist'))
+
+        name_input = self.selenium.find_element_by_id('id_form-0-name')
+        name_input.clear()
+        name_input.send_keys('other name')
+        Select(
+            self.selenium.find_element_by_name('action')
+        ).select_by_value('delete_selected')
+        self.selenium.find_element_by_name('_save').click()
+        alert = self.selenium.switch_to.alert
+        try:
+            self.assertEqual(
+                alert.text,
+                'You have selected an action, but you haven’t saved your '
+                'changes to individual fields yet. Please click OK to save. '
+                'You’ll need to re-run the action.',
+            )
+        finally:
+            alert.dismiss()
+
+    def test_save_without_changes_warns_on_pending_action(self):
+        from selenium.webdriver.support.ui import Select
+
+        Parent.objects.create(name='parent')
+
+        self.admin_login(username='super', password='secret')
+        self.selenium.get(self.live_server_url + reverse('admin:admin_changelist_parent_changelist'))
+
+        Select(
+            self.selenium.find_element_by_name('action')
+        ).select_by_value('delete_selected')
+        self.selenium.find_element_by_name('_save').click()
+        alert = self.selenium.switch_to.alert
+        try:
+            self.assertEqual(
+                alert.text,
+                'You have selected an action, and you haven’t made any '
+                'changes on individual fields. You’re probably looking for '
+                'the Go button rather than the Save button.',
+            )
+        finally:
+            alert.dismiss()
