@@ -660,14 +660,16 @@ class PluggableSearchPersonAdmin(admin.ModelAdmin):
     search_fields = ('name',)
 
     def get_search_results(self, request, queryset, search_term):
-        queryset, use_distinct = super().get_search_results(request, queryset, search_term)
+        queryset, may_have_duplicates = super().get_search_results(
+            request, queryset, search_term,
+        )
         try:
             search_term_as_int = int(search_term)
         except ValueError:
             pass
         else:
             queryset |= self.model.objects.filter(age=search_term_as_int)
-        return queryset, use_distinct
+        return queryset, may_have_duplicates
 
 
 class AlbumAdmin(admin.ModelAdmin):
